@@ -1,7 +1,7 @@
 /* lr - a better recursive ls/find */
 
 /*
- * Copyright (C) 2015-2019 Leah Neukirchen <purl.org/net/chneukirchen>
+ * Copyright (C) 2015-2020 Leah Neukirchen <purl.org/net/chneukirchen>
  * Parts of code derived from musl libc, which is
  * Copyright (C) 2005-2014 Rich Felker, et al.
  *
@@ -90,6 +90,7 @@ static int Lflag;
 static int Qflag;
 static int Pflag;
 static int Uflag;
+static int Wflag;
 static int Xflag;
 static int hflag;
 static int lflag;
@@ -97,6 +98,9 @@ static int sflag;
 static int xflag;
 static char Tflag = 'T';
 
+#define COLOR_DEFAULT -1
+#define COLOR_HIDDEN -2
+
 static char *argv0;
 static char *format;
 static char *ordering;
@@ -147,7 +151,7 @@ static int uwid, gwid, fwid;
 static time_t now;
 static mode_t default_mask;
 
-// [^a-zA-Z0-9~._/-]
+/* [^a-zA-Z0-9~._/-] */
 static char rfc3986[128] = {
 	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
 	1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,
@@ -940,7 +944,7 @@ readlin(const char *p, const char *alt)
 	return b;
 }
 
-//// AA-tree implementation, adapted from https://github.com/ccxvii/minilibs
+/* AA-tree implementation, adapted from https://github.com/ccxvii/minilibs */
 struct idtree {
 	long id;
 	char *name;
@@ -1019,7 +1023,7 @@ idtree_insert(struct idtree *node, long id, char *name)
 	}
 	return idtree_make(id, name);
 }
-////
+/**/
 
 static char *
 strid(long id)
@@ -1204,7 +1208,7 @@ xattr_string(const char *f)
 		r = sizeof xattr;
 		xattr[r-1] = 0;
 	}
-	// ignoring ENOTSUP or r == 0
+	/* ignoring ENOTSUP or r == 0 */
 
 	for (i = 0; i < r; i += strlen(xattr+i) + 1)
 		if (strcmp(xattr+i, "security.capability") == 0)
@@ -1229,7 +1233,7 @@ xattr_string(const char *f)
 #else
 	static char empty[] = "";
 	(void)f;
-	return empty;           // No support for xattrs on this platform.
+	return empty;           /* No support for xattrs on this platform. */
 #endif
 }
 
@@ -1390,7 +1394,7 @@ dircmp(char *a, char *b)
 	return *a - *b;
 }
 
-// taken straight from musl@a593414
+/* taken straight from musl@a593414 */
 int mystrverscmp(const char *l0, const char *r0)
 {
 	const unsigned char *l = (const void *)l0;
@@ -1436,7 +1440,7 @@ order(const void *a, const void *b)
 
 	for (s = ordering; *s; s++) {
 		switch (*s) {
-		// XXX use nanosecond timestamps
+		/* XXX use nanosecond timestamps */
 		case 'c': CMP(fa->sb.st_ctime, fb->sb.st_ctime);
 		case 'C': CMP(fb->sb.st_ctime, fa->sb.st_ctime);
 		case 'a': CMP(fa->sb.st_atime, fb->sb.st_atime);
@@ -1476,7 +1480,7 @@ free_fi(struct fileinfo *fi)
 	free(fi);
 }
 
-//// AA-tree implementation, adapted from https://github.com/ccxvii/minilibs
+/* AA-tree implementation, adapted from https://github.com/ccxvii/minilibs */
 struct fitree {
 	struct fileinfo *fi;
 	struct fitree *left, *right;
@@ -1560,7 +1564,7 @@ fitree_free(struct fitree *node)
 		free(node);
 	}
 }
-////
+/**/
 
 static int
 intlen(intmax_t i)
@@ -1671,11 +1675,12 @@ print_human(intmax_t i)
 
 }
 
-// Decode one UTF-8 codepoint into cp, return number of bytes to next one.
-// On invalid UTF-8, return -1, and do not change cp.
-// Invalid codepoints are not checked.
-//
-// This code is meant to be inlined, if cp is unused it can be optimized away.
+/* Decode one UTF-8 codepoint into cp, return number of bytes to next one.
+ * On invalid UTF-8, return -1, and do not change cp.
+ * Invalid codepoints are not checked.
+ *
+ * This code is meant to be inlined, if cp is unused it can be optimized away.
+ */
 static int
 u8decode(const char *cs, uint32_t *cp)
 {
@@ -1683,16 +1688,16 @@ u8decode(const char *cs, uint32_t *cp)
 
 	if (*s == 0)   { *cp = 0; return 0; }
 	if (*s < 0x80) { *cp = *s; return 1; }
-	if (*s < 0xc2) { return -1; }  //cont+overlong
+	if (*s < 0xc2) { return -1; }  /*cont+overlong*/
 	if (*s < 0xe0) { *cp = *s & 0x1f; goto u2; }
 	if (*s < 0xf0) {
-		if (*s == 0xe0 && (s[1] & 0xe0) == 0x80) return -1; //overlong
-		if (*s == 0xed && (s[1] & 0xe0) == 0xa0) return -1; //surrogate
+		if (*s == 0xe0 && (s[1] & 0xe0) == 0x80) return -1; /*overlong*/
+		if (*s == 0xed && (s[1] & 0xe0) == 0xa0) return -1; /*surrogate*/
 		*cp = *s & 0x0f; goto u3;
 	}
 	if (*s < 0xf5) {
-		if (*s == 0xf0 && (s[1] & 0xf0) == 0x80) return -1; //overlong
-		if (*s == 0xf4 && (s[1] > 0x8f)) return -1; //too high
+		if (*s == 0xf0 && (s[1] & 0xf0) == 0x80) return -1; /*overlong*/
+		if (*s == 0xf4 && (s[1] > 0x8f)) return -1; /*too high*/
 		*cp = *s & 0x07; goto u4;
 	}
 	return -1;
@@ -1711,6 +1716,12 @@ print_shquoted(const char *s)
 
 	const char *t;
 	int esc = 0;
+
+	if (!Qflag) {
+		printf("%s", s);
+		return;
+	}
+
 	for (t = s; *t; ) {
 		if ((unsigned char)*t <= 32 ||
 		    strchr("`^#*[]=|\\?${}()'\"<>&;\177", *t)) {
@@ -1725,7 +1736,7 @@ print_shquoted(const char *s)
 		}
 	}
 
-	if (!Qflag || !esc) {
+	if (!esc) {
 		printf("%s", s);
 		return;
 	}
@@ -1866,7 +1877,7 @@ color_name_on(int c, const char *f, mode_t m)
 	if (!Gflag)
 		return;
 
-	if (c != -1) {
+	if (c != COLOR_DEFAULT) {
 		fg256(c);
 		return;
 	}
@@ -1899,7 +1910,7 @@ print_urlquoted(unsigned char *s)
 static void
 hyperlink_on(char *fpath)
 {
-	// OSC 8 hyperlink format
+	/* OSC 8 hyperlink format */
 	if (Xflag) {
 		printf("\033]8;;file://%s", host);
 		if (*fpath != '/') {
@@ -1918,12 +1929,16 @@ hyperlink_off()
 		printf("\033]8;;\007");
 }
 
-// unused format codes: BEHJKLNOQVWXZ achjoqrvwz
+/* unused format codes: BEHJKLNOQVWXZ achjoqrvwz */
 void
 print_format(struct fileinfo *fi)
 {
 	int c, v;
 	char *s;
+
+	if (fi->color == COLOR_HIDDEN)
+		return;
+
 	for (s = format; *s; s++) {
 		if (*s == '\\') {
 			switch (*++s) {
@@ -2138,9 +2153,14 @@ callback(const char *fpath, const struct stat *sb, int depth, ino_t entries, off
 	fi->color = current_color;
 	memcpy((char *)&fi->sb, (char *)sb, sizeof (struct stat));
 
+	prune = 0;
 	if (expr && !eval(expr, fi)) {
-		free_fi(fi);
-		return 0;
+		if (Bflag && S_ISDIR(fi->sb.st_mode) && !prune) {
+			fi->color = COLOR_HIDDEN;
+		} else {
+			free_fi(fi);
+			return 0;
+		}
 	}
 
 	if (need_xattr) {
@@ -2150,7 +2170,7 @@ callback(const char *fpath, const struct stat *sb, int depth, ino_t entries, off
 	} else
 		memset(fi->xattr, 0, sizeof fi->xattr);
 
-	if (Uflag) {
+	if (Uflag || Wflag) {
 		print_format(fi);
 		free_fi(fi);
 		return 0;
@@ -2164,7 +2184,7 @@ callback(const char *fpath, const struct stat *sb, int depth, ino_t entries, off
 			return 0;
 		}
 	} else {
-		// add to the tree, note that this will eliminate duplicate files
+		/* add to the tree, note that this will eliminate duplicate files */
 		root = fitree_insert(root, fi);
 	}
 
@@ -2201,7 +2221,7 @@ callback(const char *fpath, const struct stat *sb, int depth, ino_t entries, off
 	return 0;
 }
 
-// lifted from musl nftw.
+/* lifted from musl nftw. */
 struct history {
 	struct history *chain;
 	dev_t dev;
@@ -2210,6 +2230,19 @@ struct history {
 	off_t total;
 };
 
+struct names {
+	char *path;
+	int guessdir;
+};
+
+int cmpstr(void const *a, void const *b) {
+    struct names *aa = (struct names *)a;
+    struct names *bb = (struct names *)b;
+
+    return strcmp(aa->path, bb->path);
+}
+
+/* NB: path is expected to have MAXPATHLEN space used in recursive calls. */
 static int
 recurse(char *path, struct history *h, int guessdir)
 {
@@ -2219,6 +2252,8 @@ recurse(char *path, struct history *h, int guessdir)
 	int r;
 	ino_t entries;
 	const char *fpath = *path ? path : ".";
+	struct names *names = 0, *tmp;
+	size_t len = 0;
 
 	int resolve = Lflag || (Hflag && !h);
 	int root = (path[0] == '/' && path[1] == 0);
@@ -2250,6 +2285,7 @@ recurse(char *path, struct history *h, int guessdir)
 
 	new.chain = h;
 	new.level = h ? h->level + 1 : 0;
+	new.total = 0;
 	if (guessdir) {
 		new.dev = st.st_dev;
 		new.ino = st.st_ino;
@@ -2258,7 +2294,6 @@ recurse(char *path, struct history *h, int guessdir)
 	entries = 0;
 
 	if (!Dflag) {
-		prune = 0;
 		r = callback(path, &st, new.level, 0, 0);
 		if (prune)
 			return 0;
@@ -2301,7 +2336,20 @@ recurse(char *path, struct history *h, int guessdir)
 #else
 				int guesssubdir = 1;
 #endif
-				if ((r = recurse(path, &new, guesssubdir))) {
+				if (Wflag) {
+					/* store paths in names for later sorting */
+					if (entries >= len) {
+						len = 2 * len + 1;
+						if (len > SIZE_MAX/sizeof *names)
+							break;
+						tmp = realloc(names, len * sizeof *names);
+						if (!tmp)
+							break;
+						names = tmp;
+					}
+					names[entries-1].path = strdup(path);
+					names[entries-1].guessdir = guesssubdir;
+				} else if ((r = recurse(path, &new, guesssubdir))) {
 					closedir(d);
 					return r;
 				}
@@ -2312,6 +2360,22 @@ recurse(char *path, struct history *h, int guessdir)
 		}
 	}
 
+	if (Wflag && names) {
+		size_t i;
+
+		qsort(names, entries, sizeof *names, cmpstr);
+
+		for (i = 0; i < entries; i++) {
+			strcpy(path, names[i].path);
+			recurse(path, &new, names[i].guessdir);
+		}
+
+		/* ensure cleanup in reverse allocation order */
+		for (i = 0; i < entries; i++)
+			free(names[entries-i-1].path);
+		free(names);
+	}
+
 	path[l] = 0;
 	if (Dflag && (r = callback(path, &st, new.level, entries, new.total)))
 		return r;
@@ -2350,7 +2414,7 @@ traverse_file(FILE *file)
 			break;
 		}
 
-		if (rd > 0 && line[rd-1] == input_delim)  // strip delimiter
+		if (rd > 0 && line[rd-1] == input_delim)  /* strip delimiter */
 			line[rd-1] = 0;
 
 		if (Lflag ? stat(line, &st) : lstat(line, &st) < 0)
@@ -2414,11 +2478,11 @@ main(int argc, char *argv[])
 
 	setlocale(LC_ALL, "");
 
-	while ((c = getopt(argc, argv, "01ABC:DFGHLQPST:UXde:f:lho:st:x")) != -1)
+	while ((c = getopt(argc, argv, "01ABC:DFGHLQPST:UWXde:f:lho:st:x")) != -1)
 		switch (c) {
 		case '0': format = zero_format; input_delim = 0; Qflag = Pflag = 0; break;
 		case '1': expr = chain(parse_expr("depth > 0 ? prune : print"), EXPR_AND, expr); break;
-		case 'A': expr = chain(expr, EXPR_AND, parse_expr("name =~ \"^\\.\" && path != \".\" ? prune : print")); break;
+		case 'A': expr = chain(expr, EXPR_AND, parse_expr("name =~ \"^\\.\" && path != \".\" ? (prune && skip) : print")); break;
 		case 'B': Bflag++; Dflag = 0; Uflag = 0; need_stat++; break;
 		case 'C':
 			if ((unsigned int)Cflag <
@@ -2428,7 +2492,7 @@ main(int argc, char *argv[])
 				fprintf(stderr, "%s: too many -C\n", argv0);
 				exit(111);
 			}
-			Gflag += 2;  // force color on
+			Gflag += 2;  /* force color on */
 			break;
 		case 'D': Dflag++; Bflag = 0; break;
 		case 'F': format = type_format; break;
@@ -2439,7 +2503,8 @@ main(int argc, char *argv[])
 		case 'P': Pflag++; Qflag++; break;
 		case 'S': Qflag++; format = stat_format; break;
 		case 'T': Tflag = timeflag(optarg); break;
-		case 'U': Uflag++; Bflag = 0; break;
+		case 'W': Wflag++; Bflag = Uflag = 0; break;
+		case 'U': Uflag++; Bflag = Wflag = 0; break;
 		case 'X': Xflag++; break;
 		case 'd': expr = chain(parse_expr("type == d && prune || print"), EXPR_AND, expr); break;
 		case 'e': expr = chain(expr, EXPR_AND,
@@ -2447,16 +2512,16 @@ main(int argc, char *argv[])
 		case 'f': format = optarg; break;
 		case 'h': hflag++; break;
 		case 'l': lflag++; Qflag++; format = long_format; break;
-		case 'o': ordering = optarg; break;
+		case 'o': Uflag = Wflag = 0; ordering = optarg; break;
 		case 's': sflag++; break;
 		case 't':
-			need_stat++;  // overapproximation
+			need_stat++;  /* overapproximation */
 			expr = chain(expr, EXPR_AND, parse_expr(optarg)); break;
 		case 'x': xflag++; break;
 		default:
 			fprintf(stderr,
 "Usage: %s [-0|-F|-l [-TA|-TC|-TM]|-S|-f FMT] [-B|-D] [-H|-L] [-1AGPQdhsx]\n"
-"          [-U|-o ORD] [-e REGEX]* [-t TEST]* [-C [COLOR:]PATH]* PATH...\n", argv0);
+"          [-U|-W|-o ORD] [-e REGEX]* [-t TEST]* [-C [COLOR:]PATH]* PATH...\n", argv0);
 			exit(2);
 		}
 
@@ -2472,7 +2537,7 @@ main(int argc, char *argv[])
 		Gflag = 0;
 
 	analyze_format();
-	if (Uflag) {
+	if (Uflag || Wflag) {
 		maxnlink = 99;
 		maxsize = 4*1024*1024;
 		maxblocks = maxsize / 512;
@@ -2492,10 +2557,10 @@ main(int argc, char *argv[])
 		}
 
 		if (gethostname(host, sizeof host) == 0) {
-			// termination not posix guaranteed
+			/* termination not posix guaranteed */
 			host[sizeof host - 1] = 0;
 		} else {
-			// ignore errors, use no host
+			/* ignore errors, use no host */
 			*host = 0;
 		}
 	}
@@ -2513,7 +2578,7 @@ main(int argc, char *argv[])
 		}
 	}
 
-	current_color = -1;
+	current_color = COLOR_DEFAULT;
 
 	initial = 1;
 	if (!Cflag && optind == argc) {
@@ -2535,7 +2600,7 @@ main(int argc, char *argv[])
 		}
 	} else if (!Uflag) {
 		fitree_walk(root, print_format);
-		// no need to destroy here, we are done
+		/* no need to destroy here, we are done */
 	}
 
 	return status;
